;RUNFROMMENU EQU 1

; Tell the AS assembler we want to code for NGPC
;
	cpu	93C141
	maxmode	on
	include	"stddef96.inc"

	include	"hardware.inc"  ; Include hardware defines
	include	"system.inc"    ; Include System Call defines

;
; Define variables
;

COL0	EQU	0
COL1	EQU	2
COL2	EQU	4
COL3	EQU	6

START_OF_RAM	EVAL	_MAINRAM
		RAMDB	VBCOUNTER
		RAMDD	COLOR
		RAMDD	OFFSET
		RAMDD	PALDST
		RAMBUF	BACKBUF,4096
		RAMBUF	LINE1,128
		RAMBUF	LINE2,128
		RAMDB	EXCH
		RAMDB	SEMITRANS

		RAMDB	CPT

	IFDEF RUNFROMMENU
HEADER_OFS      EQU     0200000H
	ELSE
HEADER_OFS      EQU     0200000H
	ENDIF


;
; Standard cartridge header
;
	org	HEADER_OFS
	db	" LICENSED BY SNK CORPORATION"	; 28 bytes license string
	dd	_start				; Program Counter
	dw	0				; Catalog number
	db	0				; Sub catalog number
	db	10h				; colour or b+w (10h = colour)
	;	 123456789012
    	db	"3D ENGINE   "			; Game name (12 bytes)
	dd	0,0,0,0				; padding - reserve for future use

; User Interrupt Vectors

UserIntVect
	dd	nada		; Software Interrupt (SWI 3)
	dd	nada		; Software Interrupt (SWI 4)
	dd	nada		; Software Interrupt (SWI 5)
	dd	nada		; Software Interrupt (SWI 6)
	dd	nada		; RTC Alarm Interrupt
	dd	VBlankInt	; Vertical Blanking Interrupt
	dd	nada		; Interrupt from Z80
	dd	nada		; Timer Interrupt (8 bit timer 0)
	dd	nada		; Timer Interrupt (8 bit timer 1)
	dd	nada		; Timer Interrupt (8 bit timer 2)
	dd	nada		; Timer Interrupt (8 bit timer 3)
	dd	nada		; Serial Transmission Interrupt
	dd	nada		; Serial Reception Interrupt
	dd	nada		; (Reserved)
	dd	nada		; End Micro DMA Int (MicroDMA 0)
	dd	nada		; End Micro DMA Int (MicroDMA 1)
	dd	nada		; End Micro DMA Int (MicroDMA 2)
	dd	nada		; End Micro DMA Int (MicroDMA 3)
nada	reti

;************************************************
; *** Start of User Code ***

_start
	calr	OS_VERSION      ; Initialize NGP or NGPC mode
	set	6,(rUSERA)      ; User Answer

; Install User Interrupt Vectors

	lda	xix,(UserIntVect)
	lda	xiy,(rSWI3)
	ld	b,18
UIVloop
        ld	xwa,(xix+)
        ld	(xiy+),xwa
        djnz	b,UIVloop

;set up screen size 160x152
	ld	(08002h),0
	ld	(08003h),0
	ld	(08004h),0a0h
	ld	(08005h),98h

; enable interrupts
	ei

;seems to be BGR format?
	ldw	(_BGCPAL),0; 0311h	;BACKGROUND COLOUR
	ld	(8118h),80h

; copy internal charset to character RAM
;	ld	ra3,3
;	ld	rw3,VECT_SYSFONTSET
;	calr	SYSTEM_CALL

	calr	InitialiseMemory

	ld	bc,(ENDPALETTE-PALETTE)/2
	ld	xde,_SCR1PAL
	ld	xhl,PALETTE
	ldirw	(xde+),(xhl+)
	ld	bc,(ENDPALETTE-PALETTE)/2
	ld	xde,_SCR2PAL
	ld	xhl,PALETTE
	ldirw	(xde+),(xhl+)

	ld	xhl,_SCR1RAM
	ld	wa,00200h
	ld	xix,_SCR2RAM
	ld	iy,00300h
	ld	de,0
	ld	b,16
fill_rows
	ld	c,16
fill_cols
	ld	(xhl+),wa
	inc	wa
	ld	(xix+),iy
	inc	iy
	dec	c
	jr	nz,fill_cols
	ld	c,16
del_cols
	ld	(xhl+),de
	ld	(xix+),de
	dec	c
	jr	nz,del_cols
	dec	b
	jr	nz,fill_rows

;*********
;Main loop
;*********

_init
	calr	clear_buf
	ld	xwa,00000h
	ld	(OFFSET),xwa
	ld	xwa,COL1
	ld	(COLOR),xwa
	ld	xwa,_SCR1PAL+8
	ld	(PALDST),xwa
	ldb	(rS1SOX),-16
	ldb	(rS2SOX),-16
	ldb	(rS1SOY),-12
	ldb	(rS2SOY),-12
	xorb	(rPF),080h

_main
;****************
; draw_hline test
;	ld	b,64
;	ld	w,0
;	ld	a,127
;_y
;	calr	draw_hline
;	inc	w
;	dec	a
;	djnz	b,_y
;****************

;***************
; draw_line test
;	ld	xhl,LINES
;_fill_buf
;	ld	wa,(xhl+)
;	ld	bc,(xhl+)
;	and	wa,07f7fh
;	and	bc,07f7fh
;	ld	xde,draw_pixel
;	calr	draw_line
;	cp	xhl,END_LINES
;	jr	nz,_fill_buf
;***************

;*******************
; draw_triangle test

	ld	xhl,LINES
_fill_buf
	ld	wa,(xhl+)
	ld	bc,(xhl+)
	ld	de,(xhl+)
	and	wa,07f7fh
	and	bc,07f7fh
	and	de,07f7fh
	calr	draw_triangle

	calr	chg_color
	xorb	(SEMITRANS),1

	push	de
	ld	xde,draw_pixel
	calr	draw_line
	pop	de
	
	ex	bc,de
	push	de
	ld	xde,draw_pixel
	calr	draw_line
	pop	de

	ex	wa,de
	push	de
	ld	xde,draw_pixel
	calr	draw_line
	pop	de

	cp	xhl,END_LINES
	jr	lt,_fill_buf

;*******************

	calr	flip_buf

	ld	w,50
	;calr	VBWait		;wait until 1 full screendraw before looping (else screen tears)

; User shutdown? (Power off pressed?)
	cp	(rUSERS),0
	jp	z,_main		;no,jump to _main and loop

; Power off NGP
	ld	rw3,VECT_SHUTDOWN
	calr	SYSTEM_CALL
_end
	jr	_end

chg_color
	push	xhl
	; change color
	ld	xhl,(COLOR)
	add	xhl,2
	cp	xhl,COL3
	jr	le,_cont0
	ld	xhl,COL1
_cont0
	ld	(COLOR),xhl
	pop	xhl
	ret

;*************************************
; VBWait
; Waits for n vertical blank interrupts
; inputs w = number of VBIs to wait for
;*************************************

VBWait
	push	xwa
	push	xbc
	ld	(VBCOUNTER),0
	ld	c,w
vbw1
	ld	w,c
	ld	a,(VBCOUNTER)
	sub	w,a
	jr	nc,vbw1
	pop	xbc
	pop	xwa
	ret

;*************************
; VBlankInt
; Vertical blank interrupt
;*************************

VBlankInt
	; Update Watch Dog Timer to prevent CPU reset
	ld	(rWDCR),WD_CLR
	; increment a counter
	inc	(VBCOUNTER)
	reti

;*************************************
; InitialiseMemory
; Initialise Memory before game starts
;*************************************

InitialiseMemory
	ld	xwa,0
	calr	ClearSprite

	; Clear Scroll 1 & 2 VRAM with 0
ClearScreen
	ld	xhl,_SCR1RAM
	ld	xwa,0
	ld	bc,1024
_clear_1
	ld	(xhl+),xwa
	djnz	bc,_clear_1

	ld	xhl,_TILERAM
	ld	xwa,0
	ld	bc,2048
_clear_2
	ld	(xhl+),xwa
	djnz	bc,_clear_2
	ret

	; Clear Sprite Attribute Memory
ClearSprite
	push	xbc
	push	xde
	push	xix
	ld	xde,0
	ld	b,64
	sll	2,wa
	ld	xix,_SPRRAM
	add	xix,xwa
SAMloop
	ld	(xix+),xde
	djnz	b,SAMloop
	pop	xix
	pop	xde
	pop	xbc
	ret

;***************************************************
; flip_buf
; Fill Tileram with back buffer & flip scroll planes 
; Level 1
;***************************************************

flip_buf
	incf

	ld	xbc,2048
	ld	xde,_TILERAM
	ld	xwa,(OFFSET)
	add	xde,xwa
	xor	xwa,01000h
	ld	(OFFSET),xwa
	ld	xhl,BACKBUF
	ldirw	(xde+),(xhl+)

	calr	clear_buf

_wait_raster
	ldb	a,(rRASV)
	and	a,07fh
	jr	z,_wait_raster

	; flip palettes
	ld	xwa,(PALDST)
	ld	xde,xwa
	xor	xwa,0180h
	ld	(PALDST),xwa

	ld	xhl,PALETTE_COL
	ld	xbc,(xhl+)
	ld	(xde+),xbc
	ld	xbc,(xhl)
	ld	(xde),xbc

	xorb	(rPF),080h

	ld	xbc,0
	ld	(xwa+),xbc
	ld	(xwa),xbc

	decf
	ret

;*************************************
; clear_buf
; Clear back buffer
;*************************************

clear_buf
	ld	xhl,BACKBUF
	ld	xwa,0
	ld	bc,4096/32
_clear_buf
	ld	(xhl+),xwa
	ld	(xhl+),xwa
	ld	(xhl+),xwa
	ld	(xhl+),xwa
	ld	(xhl+),xwa
	ld	(xhl+),xwa
	ld	(xhl+),xwa
	ld	(xhl+),xwa
	djnz	bc,_clear_buf
	ret

;*************************************
; draw_line
; Level 1
;	w0	x0
;	a0	y0
;	b0	x1
;	c0	y1
;	xde0	pixel management routine
;*************************************

draw_line
	incf
	ld	wa,rwa0
	ld	bc,rbc0

	ldb	(EXCH),0
	cp	w,b
	jr	lt,_draw_line_0
	ld	d,w
	sub	d,b
	jr	_draw_line_1
_draw_line_0
	ld	d,b
	sub	d,w
_draw_line_1
	cp	a,c
	jr	lt,_draw_line_2
	ld	e,a
	sub	e,c
	jr	_draw_line_3
_draw_line_2
	ld	e,c
	sub	e,a
_draw_line_3
	cp	d,e
	jr	ge,_draw_line_4
	ex	w,a
	ex	b,c
	ldb	(EXCH),1
_draw_line_4

	cp	w,b
	jr	lt,_draw_line
	ex	wa,bc
_draw_line

	; dx
	ld	de,0
	ld	hl,0
	ld	e,b
	ld	l,w
	sub	de,hl
	ld	ix,de

	; dy
	ld	de,0
	ld	hl,0
	ld	e,c
	ld	l,a
	cp	l,e
	jr	le,_inc
	ld	c,-1
	sub	hl,de
	ld	de,hl
	jr	_dy
_inc
	ld	c,1
	sub	de,hl
_dy
	ld	iy,de

	; d, incre, incrne
	add	de,iy
	ld	iz,de	; incre
	sub	de,ix	; d
	ld	hl,de
	sub	hl,ix	; incrne

cont_drawline
	call	(xde0)
	cp	w,b
	jr	z,end_drawline
	inc	1,w
	cp	de,0
	jr	gt,_cont1_drawline
	add	de,iz
	jr	cont_drawline
_cont1_drawline
	add	de,hl
	add	a,c
	jr	cont_drawline
end_drawline
	; end
	decf
	ret

;*************************************
; draw_pixel
; Level 2
;	w1	x
;	a1	y
;*************************************

draw_pixel
	incf

	ld	wa,rwa1
	cpb	(EXCH),0
	jr	z,_no_exch
	ex	w,a
_no_exch
	ld	xhl,0
	ld	xbc,0
	ld	l,w
	ld	c,w
	srl	3,c
	sll	4,c
	ld	b,a
	srl	3,b
	and	a,7
	sll	1,a
	ld	w,0
	add	bc,wa
	and	l,7
	sll	1,hl
	add	xhl,mask_pixel
	ld	hl,(xhl)
	ld	xde,set_pixel
	add	xde,(COLOR)
	ld	wa,(xde)
	add	xbc,BACKBUF
	and	(xbc),hl
	xor	hl,0ffffh
	and	wa,hl
	or	(xbc),wa

	decf
	ret

;*************************************
; draw_hline
; Draw horizontal line
; Level 1
;	w0	x0
;	a0	x1
;	b0	y
;*************************************

draw_hline
	incf

	ld	wa,rwa0
	cp	w,a
	jr	le,_draw_hline
	ex	w,a
_draw_hline
	xor	xbc,xbc
	xor	xde,xde
	ld	c,rb0
	ld	d,c
	srl	3,d
	and	c,7
	sll	1,c
	add	de,bc
	xor	bc,bc
	ld	c,w
	and	c,0f8h
	sll	1,c
	add	de,bc

	xor	hl,hl
	ld	l,a
	and	l,0f8h
	sll	1,l

	add	xde,BACKBUF
	and	w,7
	and	a,7

	push	xix
	ld	xix,set_pixel
	add	xix,(COLOR)
	ld	ix,(xix)
	push	iz
	xor	iz,iz
	cpb	(SEMITRANS),0
	jr	z,_no_trans
	ld	iz,0cccch
	bit	0,rb0
	jr	z,_no_trans
	ld	iz,03333h
_no_trans
	sub	hl,bc
	jr	z,_same_addr

	; not same addr : loop from left to right

	push	xiy
	ld	iy,hl

	; draw left side
	xor	xhl,xhl
	ld	l,w
	sll	1,l
	add	xhl,mask_left
	ld	bc,(xhl)
	or	bc,iz
	and	(xde),bc
	xor	bc,0ffffh
	and	bc,ix
	or	(xde),bc
	add	xde,16

	; draw middle
_draw_middle
	cp	iy,16
	jr	le,_draw_right
	ld	bc,iz
	and	(xde),bc
	xor	bc,0ffffh
	and	bc,ix
	or	(xde),bc
	add	xde,16
	sub	iy,16
	jr	_draw_middle

	; draw right side
_draw_right
	xor	xhl,xhl
	ld	l,a
	sll	1,l
	add	xhl,mask_right
	ld	bc,(xhl)
	or	bc,iz
	and	(xde),bc
	xor	bc,0ffffh
	and	bc,ix
	or	(xde),bc

	pop	xiy
	pop	iz
	pop	xix
	decf
	ret

_same_addr
	xor	xhl,xhl
	ld	l,w
	sll	1,l
	add	xhl,mask_left
	ld	bc,(xhl)
	xor	xhl,xhl
	ld	l,a
	sll	1,l
	add	xhl,mask_right
	or	bc,(xhl)
	or	bc,iz
	and	(xde),bc
	xor	bc,0ffffh
	and	bc,ix
	or	(xde),bc

	pop	iz
	pop	xix
	decf
	ret

set_pixel
	dw	00000h,05555h,0aaaah,0ffffh
mask_pixel
	dw	03fffh,0cfffh,0f3ffh,0fcffh,0ff3fh,0ffcfh,0fff3h,0fffch
mask_left
	dw	00000h,0c000h,0f000h,0fc00h,0ff00h,0ffc0h,0fff0h,0fffch
mask_right
	dw	03fffh,00fffh,003ffh,000ffh,0003fh,0000fh,00003h,00000h


;*************************************
; draw_triangle
; Draw triangle
; Level 0
;	wa0	x0y0
;	bc0	x1y1
;	de0	x2y2
;*************************************

draw_triangle
	push	xwa
	push	xbc
	push	xde
	push	xhl

	cp	a,c
	jr	le,_a_inf_c
	ex	wa,bc
_a_inf_c
	cp	a,e
	jr	le,_a_inf_e
	ex	wa,de
_a_inf_e
	cp	c,e
	jr	le,_c_inf_e
	ex	de,bc
_c_inf_e
	; wa : higher point
	; bc : mid point
	; de : lower point

	; fill triangle : wa,bc,dc
	cp	b,d
	jr	gt,_inv_1
	jr	nz,_no_inv_1
	cp	b,w		; when b==d
	jr	gt,_inv_1
_no_inv_1
	calr	_comp_line_left
	ex	bc,de
	calr	_comp_line_right
	ex	bc,de
	jr	_draw_first
_inv_1
	calr	_comp_line_right
	ex	bc,de
	calr	_comp_line_left
	ex	bc,de

_draw_first
	push	wa
	push	bc
	push	de
	push	xhl	; address of last point on wa -> de at y = c
	ld	b,a
	sub	c,a
	inc	c
	xor	xde,xde
	ld	e,b
	add	xde,LINE1
	xor	xhl,xhl
	ld	l,b
	add	xhl,LINE2
_fill_1
	ld	w,(xde+)
	ld	a,(xhl+)
	calr	draw_hline
	inc	b
	djnz	c,_fill_1
	pop	xhl
	pop	de
	pop	bc
	pop	wa

	; fill triangle : bc,d'c,de
	ex	wa,de
	cp	b,w
	jr	le,_inv_2
_do_right
	calr	_comp_line_right
	jr	_draw_second
_inv_2
	jr	nz,_inv_3
	cp	b,d
	jr	gt,_do_right
_inv_3	
	calr	_comp_line_left

_draw_second
	ld	b,c
	ex	c,a
	sub	c,a
	inc	c
	xor	xde,xde
	ld	e,b
	add	xde,LINE1
	xor	xhl,xhl
	ld	l,b
	add	xhl,LINE2
_fill_2
	ld	w,(xde+)
	ld	a,(xhl+)
	calr	draw_hline
	inc	b
	djnz	c,_fill_2

	pop	xhl
	pop	xde
	pop	xbc
	pop	xwa
	ret

	;***********
	; Left line
	;***********
_comp_line_left
	push	xde
	push	bc
	ld	xde,LINE1
	ld	b,127
	ld	c,128
_fill_line1
	ld	(xde+),b
	djnz	c,_fill_line1
	pop	bc
	ld	xde,draw_left
	calr	draw_line
	pop	xde
	ret

	;************
	; Right line
	;************
_comp_line_right
	push	xde
	push	bc
	ld	xde,LINE2
	ld	b,0
	ld	c,128
_fill_line2
	ld	(xde+),b
	djnz	c,_fill_line2
	pop	bc
	ld	xde,draw_right
	calr	draw_line
	pop	xde
	ret

;**************************
; Draw left pixel on LINE1
;**************************
draw_left
	incf
	ld	wa,rwa1
	cpb	(EXCH),0
	jr	z,_no_exch_left
	ex	w,a
_no_exch_left
	xor	xhl,xhl
	ld	l,a
	add	xhl,LINE1
	cp	w,(xhl)
	jr	ge,_end_draw_left
	ld	(xhl),w
_end_draw_left
	decf
	ret

;***************************
; Draw right pixel on LINE2
;***************************

draw_right
	incf
	ld	wa,rwa1
	cpb	(EXCH),0
	jr	z,_no_exch_right
	ex	w,a
_no_exch_right
	xor	xhl,xhl
	ld	l,a
	add	xhl,LINE2
	cp	w,(xhl)
	jr	lt,_end_draw_right
	ld	(xhl),w
_end_draw_right
	decf
	ret


;*****************************************************************************


PALETTE
	dw 00fffh, 00fffh, 0fffh, 00fffh	;black
PALETTE_COL
	dw 00000h, 00f00h, 00f0h, 0000fh
ENDPALETTE

LINES
	db 68, 65, 106, 177, 141, 95, 175, 197, 206, 219, 44, 19, 183, 57, 163, 76
	db 210, 142, 180, 144, 23, 140, 62, 137, 9, 90, 32, 198, 239, 90, 160, 188
	db 128, 34, 10, 91, 25, 151, 214, 167, 116, 217, 200, 179, 73, 19, 204, 12
	db 97, 233, 164, 220, 46, 73, 86, 144, 132, 26, 159, 128, 0, 15, 208, 23
	db 81, 233, 186, 154, 182, 135, 89, 130, 190, 24, 135, 142, 144, 48, 130, 58
	db 83, 41, 218, 120, 51, 168, 239, 204, 41, 162, 101, 53, 165, 24, 40, 83
	db 66, 251, 4, 173, 178, 169, 160, 176, 206, 124, 144, 121, 54, 49, 228, 21
	db 224, 217, 192, 170, 175, 99, 1, 188, 204, 112, 243, 128, 127, 6, 74, 209
	db 67, 221, 72, 210, 170, 24, 5, 201, 189, 143, 166, 172, 138, 201, 192, 226
	db 140, 53, 250, 11, 208, 115, 102, 120, 113, 245, 74, 151, 237, 110, 72, 69
	db 150, 17, 184, 128, 253, 103, 46, 214, 56, 211, 104, 99, 5, 41, 74, 46
	db 102, 218, 212, 175, 120, 115, 65, 164, 203, 159, 177, 154, 63, 49, 102, 145
	db 143, 77, 204, 147, 32, 183, 142, 49, 133, 27, 195, 134, 201, 58, 16, 232
	db 235, 171, 224, 122, 252, 122, 40, 110, 118, 163, 107, 191, 46, 81, 86, 195
	db 53, 39, 61, 234, 222, 113, 239, 169, 144, 105, 252, 101, 55, 57, 32, 217
	db 131, 210, 229, 155, 137, 235, 41, 121, 235, 18, 57, 182, 35, 227, 94, 42
END_LINES

SYSTEM_CALL
	SystemCallCode
OS_VERSION
	OsVersionCode
